from bloodyAD.formatters import (
    accesscontrol,
    common,
    cryptography,
    dns,
)
from bloodyAD.exceptions import LOG
import base64
from functools import lru_cache
from winacl.dtyp.security_descriptor import SECURITY_DESCRIPTOR


def formatAccountControl(userAccountControl):
    userAccountControl = int(userAccountControl.decode())
    return [
        key
        for key, val in accesscontrol.ACCOUNT_FLAGS.items()
        if userAccountControl & val == val
    ]


def formatTrustDirection(trustDirection):
    trustDirection = int(trustDirection.decode())
    for key, val in common.TRUST_DIRECTION.items():
        if trustDirection == val:
            return key
    return trustDirection


def formatTrustAttributes(trustAttributes):
    trustAttributes = int(trustAttributes.decode())
    return [
        key
        for key, val in common.TRUST_ATTRIBUTES.items()
        if trustAttributes & val == val
    ]


def formatTrustType(trustType):
    trustType = int(trustType.decode())
    for key, val in common.TRUST_TYPE.items():
        if trustType == val:
            return key
    return trustType


def formatSD(sd_bytes):
    return SECURITY_DESCRIPTOR.from_bytes(sd_bytes).to_sddl()


def formatFunctionalLevel(behavior_version):
    behavior_version = behavior_version.decode()
    return (
        common.FUNCTIONAL_LEVEL[behavior_version]
        if behavior_version in common.FUNCTIONAL_LEVEL
        else behavior_version
    )


def formatSchemaVersion(objectVersion):
    objectVersion = objectVersion.decode()
    return (
        common.SCHEMA_VERSION[objectVersion]
        if objectVersion in common.SCHEMA_VERSION
        else objectVersion
    )


def formatGMSApass(managedPassword):
    gmsa_blob = cryptography.MSDS_MANAGEDPASSWORD_BLOB(managedPassword)
    nt_hash = gmsa_blob.toNtHash()
    return {
        "NT": nt_hash,
        "B64ENCODED": base64.b64encode(gmsa_blob["CurrentPassword"]).decode(),
    }


def formatDnsRecord(dns_record):
    return dns.Record(dns_record).toDict()


def formatWellKnownObjects(wellKnown_object):
    dn_binary = common.DNBinary(wellKnown_object)
    if dn_binary.binary_value in common.WELLKNOWN_GUID:
        dn_binary.binary_value = common.WELLKNOWN_GUID[dn_binary.binary_value]
    return dn_binary


def formatKeyCredentialLink(key_dnbinary):
    return cryptography.KEYCREDENTIALLINK_BLOB(
        common.DNBinary(key_dnbinary).value
    ).toDict()


from badldap.protocol.typeconversion import (
    LDAP_WELL_KNOWN_ATTRS,
    MSLDAP_BUILTIN_ATTRIBUTE_TYPES,
    single_guid,
    int2timedelta
)


def formatFactory(format_func, origin_format):
    def genericFormat(val, encode=False, *args):
        if encode:
            return origin_format(val, encode, *args)
        if not isinstance(val, list):
            return format_func(val)
        return [format_func(e) for e in val]
    # The function name is set to the original function name for encode changes logic
    genericFormat.__name__ = origin_format.__name__
    return genericFormat

@lru_cache
def getFormatters():
    """
    Returns a dictionary mapping attribute names to their formatting functions.
    This doesn't modify badldap's global dictionaries, allowing for local formatting.
    
    Returns both bloodyAD custom formatters and badldap default formatters.
    BloodyAD formatters take precedence over badldap formatters.
    """
    def make_formatter(format_func):
        """Wrapper to handle list/non-list values consistently"""
        def wrapper(val):
            if isinstance(val, list):
                # if len(val) == 1:
                #     return format_func(val[0])
                # else:
                return [format_func(v) for v in val]
            else:
                return format_func(val)
        return wrapper
    
    def make_list_formatter(format_func):
        """Wrapper for formatters that expect the value as a list"""
        def wrapper(val):
            if isinstance(val, list):
                return format_func(val)
            else:
                return format_func([val])
        return wrapper
    
    # Start with badldap's default formatters
    formatters_map = {}
    
    # First, import badldap's MSLDAP_BUILTIN_ATTRIBUTE_TYPES formatters (higher priority)
    for attr_name, formatter in MSLDAP_BUILTIN_ATTRIBUTE_TYPES.items():
        formatters_map[attr_name] = make_list_formatter(formatter)
    
    # Then, add LDAP_WELL_KNOWN_ATTRS formatters only if not already present
    for attr_name, formatter in LDAP_WELL_KNOWN_ATTRS.items():
        if attr_name not in formatters_map:
            formatters_map[attr_name] = make_list_formatter(formatter)
    
    # Now override with bloodyAD custom formatters (these take highest precedence)
    # Security descriptors - expect single bytes value
    formatters_map["nTSecurityDescriptor"] = make_formatter(formatSD)
    formatters_map["msDS-AllowedToActOnBehalfOfOtherIdentity"] = make_formatter(formatSD)
    formatters_map["msDS-GroupMSAMembership"] = make_formatter(formatSD)
    
    # Passwords and credentials - expect single bytes value
    formatters_map["msDS-ManagedPassword"] = make_formatter(formatGMSApass)
    
    # Account control - expect single bytes value
    formatters_map["userAccountControl"] = make_formatter(formatAccountControl)
    formatters_map["msDS-User-Account-Control-Computed"] = make_formatter(formatAccountControl)
    
    # Trust attributes - expect single bytes value
    formatters_map["trustDirection"] = make_formatter(formatTrustDirection)
    formatters_map["trustAttributes"] = make_formatter(formatTrustAttributes)
    formatters_map["trustType"] = make_formatter(formatTrustType)
    
    # Versions and levels - expect single bytes value
    formatters_map["msDS-Behavior-Version"] = make_formatter(formatFunctionalLevel)
    formatters_map["objectVersion"] = make_formatter(formatSchemaVersion)
    
    # DNS and other - expect single bytes value
    formatters_map["dnsRecord"] = make_formatter(formatDnsRecord)
    formatters_map["msDS-KeyCredentialLink"] = make_formatter(formatKeyCredentialLink)
    formatters_map["wellKnownObjects"] = make_formatter(formatWellKnownObjects)
    
    formatters_map["attributeSecurityGUID"] = make_formatter(single_guid)
    formatters_map["msDS-MinimumPasswordAge"] = make_formatter(int2timedelta)
    # GUID and time attributes - keep badldap's formatters (already added above)
    # attributeSecurityGUID and msDS-MinimumPasswordAge use badldap's single_guid and int2timedelta
    
    return formatters_map


def applyFormatters(attributes):
    """
    Apply formatters to attributes dictionary.
    
    Args:
        attributes: Dictionary of attribute names to values
        formatters_map: Dictionary of attribute names to formatter functions
    
    Returns:
        Dictionary with formatted attributes
    """
    formatted_attrs = {}
    formatters_map = getFormatters()
    for attr_name, attr_value in attributes.items():
        if attr_name in formatters_map:
            formatter = formatters_map[attr_name]
            try:
                formatted_attrs[attr_name] = formatter(attr_value)
            except Exception as e:
                # If formatting fails, log the error and keep original value
                LOG.debug(
                    f"Failed to format attribute '{attr_name}': {type(e).__name__}: {e}"
                )
                formatted_attrs[attr_name] = attr_value
        else:
            formatted_attrs[attr_name] = attr_value
    
    return formatted_attrs


from winacl.dtyp.ace import (
    SYSTEM_AUDIT_OBJECT_ACE,
    SDDL_ACE_TYPE_MAPS_INV,
    aceflags_to_sddl,
    accessmask_to_sddl,
    ACE_OBJECT_PRESENCE,
)


def to_sddl(self, sd_object_type=None):
    # ace_type;ace_flags;rights;object_guid;inherit_object_guid;account_sid;(resource_attribute)
    return "(%s;%s;%s;%s;%s;%s)" % (
        SDDL_ACE_TYPE_MAPS_INV[self.AceType],
        aceflags_to_sddl(self.AceFlags),
        accessmask_to_sddl(self.Mask, self.sd_object_type),
        (
            self.ObjectType.to_bytes()
            if self.AceFlags & ACE_OBJECT_PRESENCE.ACE_OBJECT_TYPE_PRESENT
            else ""
        ),
        (
            self.InheritedObjectType.to_bytes()
            if self.Flags & ACE_OBJECT_PRESENCE.ACE_INHERITED_OBJECT_TYPE_PRESENT
            else ""
        ),
        self.Sid.to_sddl(),
    )


setattr(SYSTEM_AUDIT_OBJECT_ACE, "to_sddl", to_sddl)



import binascii
from typing import Literal
import badldap
from bloodyAD import utils, ConnectionHandler
from bloodyAD.exceptions import LOG
from bloodyAD.formatters import accesscontrol, common, dns, cryptography
from bloodyAD.exceptions import BloodyError
from bloodyAD.network.ldap import Change


async def dcsync(conn: ConnectionHandler, trustee: str):
    """
    Remove DCSync right for provided trustee

    :param trustee: sAMAccountName, DN or SID of the trustee
    """
    ldap = await conn.getLdap()
    new_sd, _ = await utils.getSD(conn, ldap.domainNC)
    if "s-1-" in trustee.lower():
        trustee_sid = trustee
    else:
        entry = None
        async for e in ldap.bloodysearch(trustee, attr=["objectSid"]):
            entry = e
            break
        trustee_sid = entry["objectSid"]
    access_mask = accesscontrol.ACCESS_FLAGS["ADS_RIGHT_DS_CONTROL_ACCESS"]
    utils.delRight(new_sd, trustee_sid, access_mask)

    req_flags = badldap.wintypes.asn1.sdflagsrequest.SDFlagsRequestValue(
        {"Flags": accesscontrol.DACL_SECURITY_INFORMATION}
    )
    controls = [("1.2.840.113556.1.4.801", True, req_flags.dump())]

    await ldap.bloodymodify(
        ldap.domainNC,
        {"nTSecurityDescriptor": [(Change.REPLACE.value, new_sd.getData())]},
        controls,
    )

    LOG.info(f"{trustee} can't DCSync anymore")


async def dnsRecord(
    conn: ConnectionHandler,
    name: str,
    data: str,
    dnstype: Literal["A", "AAAA", "CNAME", "MX", "PTR", "SRV", "TXT"] = "A",
    zone: str = "CurrentDomain",
    ttl: int = None,
    preference: int = None,
    port: int = None,
    priority: int = None,
    weight: int = None,
    forest: bool = False,
):
    """
    Remove a DNS record of an AD environment.

    :param name: name of the dnsNode object (hostname) which contains the record
    :param data: DNS record data
    :param dnstype: DNS record type
    :param zone: DNS zone
    :param ttl: DNS record TTL
    :param preference: DNS MX record preference
    :param port: listening port of the service in a DNS SRV record
    :param priority: priority of a DNS SRV record against concurrent
    :param weight: weight of a DNS SRV record against concurrent
    :param forest: if set, will fetch the dns record in forest instead of domain
    """

    ldap = await conn.getLdap()
    naming_context = "," + ldap.domainNC
    if zone == "CurrentDomain":
        zone = ""
        for label in naming_context.split(",DC="):
            if label:
                zone += "." + label
        if forest:
            zone = "_msdcs" + zone
        else:
            # Removes first dot
            zone = zone[1:]

    # TODO: take into account custom ForestDnsZones and DomainDnsZones partition name ?
    if forest:
        zone_type = "ForestDnsZones"
    else:
        zone_type = "DomainDnsZones"

    zone_dn = f",DC={zone},CN=MicrosoftDNS,DC={zone_type}{naming_context}"
    record_dn = f"DC={name}{zone_dn}"

    record_to_remove = None
    entry = None
    async for e in ldap.bloodysearch(record_dn, attr=["dnsRecord"], raw=True):
        entry = e
        break
    dns_list = entry["dnsRecord"]
    for raw_record in dns_list:
        record = dns.Record(raw_record)
        tmp_record = dns.Record()

        if not ttl:
            ttl = record["TtlSeconds"]
        tmp_record.fromDict(
            data,
            dnstype,
            ttl,
            record["Rank"],
            record["Serial"],
            preference,
            port,
            priority,
            weight,
        )
        if tmp_record.getData() == raw_record:
            record_to_remove = raw_record
            break

    if not record_to_remove:
        LOG.warning("Record not found")
        return

    if len(dns_list) > 1:
        await ldap.bloodymodify(
            record_dn, {"dnsRecord": [(Change.DELETE.value, record_to_remove)]}
        )
    else:
        await ldap.bloodydelete(record_dn)

    LOG.info(f"Given record has been successfully removed from {name}")


async def genericAll(conn: ConnectionHandler, target: str, trustee: str):
    """
    Remove full control of trustee on target

    :param target: sAMAccountName, DN or SID of the target
    :param trustee: sAMAccountName, DN or SID of the trustee
    """
    ldap = await conn.getLdap()
    new_sd, _ = await utils.getSD(conn, target)
    if "s-1-" in trustee.lower():
        trustee_sid = trustee
    else:
        entry = None
        async for e in ldap.bloodysearch(trustee, attr=["objectSid"]):
            entry = e
            break
        trustee_sid = entry["objectSid"]
    utils.delRight(new_sd, trustee_sid)

    req_flags = badldap.wintypes.asn1.sdflagsrequest.SDFlagsRequestValue(
        {"Flags": accesscontrol.DACL_SECURITY_INFORMATION}
    )
    controls = [("1.2.840.113556.1.4.801", True, req_flags.dump())]

    await ldap.bloodymodify(
        target,
        {"nTSecurityDescriptor": [(Change.REPLACE.value, new_sd.getData())]},
        controls,
    )

    LOG.info(f"{trustee} doesn't have GenericAll on {target} anymore")


async def groupMember(conn: ConnectionHandler, group: str, member: str):
    """
    Remove member (user, group, computer) from group

    :param group: sAMAccountName, DN or SID of the group
    :param member: sAMAccountName, DN or SID of the member
    """
    # This is equivalent to classic add member,
    # see [MS-ADTS] - 3.1.1.3.1.2.4 Alternative Forms of DNs
    # But <SID='sid'> also has the advantage of being compatible with foreign security principals,
    # see [MS-ADTS] - 3.1.1.5.3.3 Processing Specifics
    ldap = await conn.getLdap()
    if "s-1-" in member.lower():
        # We assume member is an SID
        member_transformed = f"<SID={member}>"
    else:
        member_transformed = await ldap.dnResolver(member)

    await ldap.bloodymodify(
        group, {"member": [(Change.DELETE.value, member_transformed)]}
    )
    LOG.info(f"{member} removed from {group}")


async def object(conn: ConnectionHandler, target: str):
    """
    Remove object (user, group, computer, organizational unit, etc)

    :param target: sAMAccountName, DN or SID of the target
    """
    ldap = await conn.getLdap()
    await ldap.bloodydelete(target)
    LOG.info(f"{target} has been removed")


async def rbcd(conn: ConnectionHandler, target: str, service: str):
    """
    Remove Resource Based Constraint Delegation for service on target

    :param target: sAMAccountName, DN or SID of the target
    :param service: sAMAccountName, DN or SID of the service account
    """
    ldap = await conn.getLdap()
    control_flag = 0
    new_sd, _ = await utils.getSD(
        conn, target, "msDS-AllowedToActOnBehalfOfOtherIdentity", control_flag
    )
    if "s-1-" in service.lower():
        service_sid = service
    else:
        entry = None
        async for e in ldap.bloodysearch(service, attr=["objectSid"]):
            entry = e
            break
        service_sid = entry["objectSid"]
    access_mask = accesscontrol.ACCESS_FLAGS["ADS_RIGHT_DS_CONTROL_ACCESS"]
    utils.delRight(new_sd, service_sid, access_mask)

    attr_values = []
    if len(new_sd["Dacl"].aces) > 0:
        attr_values = new_sd.getData()
    await ldap.bloodymodify(
        target,
        {
            "msDS-AllowedToActOnBehalfOfOtherIdentity": [
                (
                    Change.REPLACE.value,
                    attr_values,
                )
            ]
        },
    )

    LOG.info(f"{service} can't impersonate users on {target} anymore")


async def shadowCredentials(conn: ConnectionHandler, target: str, key: str = None):
    """
    Remove Key Credentials from target

    :param target: sAMAccountName, DN or SID of the target
    :param key: RSA key of Key Credentials to remove from the target, removes all if key not specified
    """
    ldap = await conn.getLdap()
    entry = None
    async for e in ldap.bloodysearch(target, attr=["msDS-KeyCredentialLink"], raw=True):
        entry = e
        break
    keyCreds = entry.get("msDS-KeyCredentialLink", [])
    newKeyCreds = []
    isFound = False
    for keyCred in keyCreds:
        key_raw = common.DNBinary(keyCred).value
        key_blob = cryptography.KEYCREDENTIALLINK_BLOB(key_raw)
        if key and key_blob.getKeyID() != binascii.unhexlify(key):
            newKeyCreds.append(keyCred.decode())
        else:
            isFound = True
            LOG.debug("Key to delete found")

    if not isFound:
        LOG.warning("No key found")
        return
       
    await ldap.bloodymodify(
        target, {"msDS-KeyCredentialLink": [(Change.REPLACE.value, newKeyCreds)]}
    )
    str_key = key if key else "All keys"
    LOG.info(f"{str_key} removed")


async def uac(conn: ConnectionHandler, target: str, f: list = None):
    """
    Remove property flags altering user/computer object behavior

    :param target: sAMAccountName, DN or SID of the target
    :param f: name of property flag to remove, can be called multiple times if multiple flags to remove (e.g -f LOCKOUT  -f ACCOUNTDISABLE)
    """
    ldap = await conn.getLdap()
    uac = 0
    for flag in f:
        uac |= accesscontrol.ACCOUNT_FLAGS[flag]

    try:
        entry = None
        async for e in ldap.bloodysearch(target, attr=["userAccountControl"], raw=True):
            entry = e
            break
        old_uac = entry["userAccountControl"][0]
    except IndexError as e:
        entry = None
        async for search_entry in ldap.bloodysearch(target, attr=["allowedAttributes"]):
            entry = search_entry
            break
        for allowed in entry["allowedAttributes"]:
            if "userAccountControl" in allowed:
                raise BloodyError(
                    "Current user doesn't have the right to read userAccountControl on"
                    f" {target}"
                ) from e
        raise BloodyError(f"{target} doesn't have userAccountControl attribute") from e

    uac = int(old_uac) & ~uac
    await ldap.bloodymodify(
        target, {"userAccountControl": [(Change.REPLACE.value, uac)]}
    )

    LOG.info(f"{f} property flags removed from {target}'s userAccountControl")



from badldap.ldap_objects import (
    MSADUser, MSADMachine, MSADGroup, MSADOU, MSADGPO, 
    MSADContainer, MSADDMSAUser, MSADGMSAUser, MSADDomainTrust
)
from badldap.wintypes.asn1.sdflagsrequest import SDFlagsRequestValue
from bloodyAD.formatters import accesscontrol
from badldap.external.bloodhoundpy.resolver import resolve_aces, WELLKNOWN_SIDS
from badldap.external.bloodhoundpy.acls import parse_binary_acl
from badldap.bloodhound import MSLDAPDump2Bloodhound
import zipfile
from bloodyAD.network.ldap import showRecoverable
from bloodyAD.exceptions import LOG
from bloodyAD.utils import global_lazy_adschema

def create_msldapentry(entry, otype):
    """
    Create a badldap ldapentry object from a dictionary entry based on resolved type.
    """
    
    if otype == 'user':
        object_class = entry["attributes"].get('objectClass', [])
        if 'msDS-GroupManagedServiceAccount' in object_class:
            return MSADGMSAUser.from_ldap(entry)
        elif 'msDS-ManagedServiceAccount' in object_class:
            return MSADDMSAUser.from_ldap(entry)
        else:
            return MSADUser.from_ldap(entry)
    elif otype == 'computer':
        return MSADMachine.from_ldap(entry)
    elif otype == 'group':
        return MSADGroup.from_ldap(entry)
    elif otype == 'gpo':
        return MSADGPO.from_ldap(entry)
    elif otype == 'ou':
        return MSADOU.from_ldap(entry)
    elif otype == 'domain':
        return MSADDomainTrust.from_ldap(entry)
    else:
        # container, domain, base, trustaccount, or unknown
        return MSADContainer.from_ldap(entry)


async def granular_bh(conn, searchbase, ldap_filter, output_path=None):
    """
    Generate BloodHound-compatible JSON for queried objects.
    """
    ldap = await conn.getLdap()
    msbh = MSLDAPDump2Bloodhound(ldap.co_url)
    additional_schema = {
        'ms-mcs-admpwd': '79775c0c-d2e0-4b5f-b8e5-0e8e6a0c0e0e',  # LAPS password attribute
        'ms-laps-encryptedpassword': 'c835d1d5-f9fb-4c5b-8b8f-8b6f9b6f9b6f',  # LAPS encrypted password
        'ms-ds-key-credential-link': '5b47d60f-6090-40b2-9f37-2a4de88f3063',  # Key Credential Link
        'service-principal-name': 'f3a64788-5306-11d1-a9c5-0000f80367c1',  # SPN attribute
        'user-principal-name': '28630ebf-41d5-11d1-a9c1-0000f80367c1',  # UPN attribute
    }
    msbh.schema.update(additional_schema)
    msbh.domainname = ldap.domainname
    bh_data = {}
    filectr_dict = {}
    group_members = {}
    adinfo, err = await ldap.get_ad_info()
    if err:
        raise err
    control_flag=(
            accesscontrol.OWNER_SECURITY_INFORMATION
            + accesscontrol.GROUP_SECURITY_INFORMATION
            + accesscontrol.DACL_SECURITY_INFORMATION
        )
    req_flags = SDFlagsRequestValue({"Flags": control_flag})
    # First one for recycled and second one for nt security descriptor
    controls = showRecoverable() +[("1.2.840.113556.1.4.801", True, req_flags.dump())]
    with zipfile.ZipFile(msbh.zipfilepath, 'w', zipfile.ZIP_DEFLATED) as msbh.zipfile:
        async for entry, err in ldap.pagedsearch(ldap_filter, attributes=['*'], tree=searchbase, controls=controls):
            if err:
                raise err
            entry_attr = entry['attributes']
            # Deal with known foreign principals
            gname = ""
            if entry_attr.get('name') in WELLKNOWN_SIDS:
                    bh_entry = {}
                    gname, sidtype = WELLKNOWN_SIDS[entry_attr['name']]
                    obj_type = sidtype.lower()
                    # bh_entry['type'] = sidtype.capitalize()
                    # bh_entry['principal'] = '%s@%s' % (gname.upper(), msbh.domainname.upper())
                    # bh_entry['ObjectIdentifier'] = '%s-%s' % (msbh.domainname.upper(), entry_attr['objectSid'].upper())
            else:
                resolved = msbh.resolve_entry(entry_attr)
                obj_type = resolved['type'].lower()
                if obj_type == 'trustaccount':
                    obj_type = 'user'  # We will consider Trust accounts are users in BH for now
            msldap_entry = create_msldapentry(entry, obj_type)                
            try:
                bh_entry = msldap_entry.to_bh(ldap.domainname, adinfo.objectSid)
            except TypeError:
                try:
                    bh_entry = msldap_entry.to_bh(ldap.domainname)
                except TypeError:
                    bh_entry = msldap_entry.to_bh()
            if gname:
                bh_entry['Properties']['name'] = gname
            elif entry_attr.get('isDeleted'):
                bh_entry['Properties']['name'] = '%s@%s' % (entry_attr['name'].replace('\n',' ').upper(), ldap.domainname.upper())
            # Parse the ACL
            dn_entry, bh_entry, relations = parse_binary_acl(
                entry_attr['distinguishedName'], 
                bh_entry, 
                obj_type, 
                entry['attributes']['nTSecurityDescriptor'], 
                msbh.schema
            )

            # Add to object cache for group membership resolution
            msbh.add_ocache(dn_entry, bh_entry['ObjectIdentifier'], bh_entry['Properties']['name'], obj_type)

            bh_entry['Aces'] = resolve_aces(relations, ldap.domainname, adinfo.objectSid, msbh.ocache)
            
            json_type = obj_type + 's'
            bh_type_data = bh_data.get(json_type, msbh.get_json_wrapper(json_type))
            bh_type_data['data'].append(bh_entry)
            bh_type_data['meta']['count'] += 1

            # For groups we need post processing to resolve members so we cannot flush now
            if json_type == 'groups':
                group_members[bh_entry['ObjectIdentifier']] = entry_attr.get('member', [])
            elif bh_type_data['meta']['count'] == msbh.MAX_ENTRIES_PER_FILE:
                LOG.info('Max entries per file reached for %s, flushing %d entries to zip' % (json_type, bh_type_data['meta']['count']))
                filectr = filectr_dict.get(json_type, 0)
                await msbh.write_json_to_zip(json_type, bh_type_data, filectr)
                bh_type_data = msbh.get_json_wrapper(json_type)
                filectr_dict[json_type] = filectr + 1
            bh_data[json_type] = bh_type_data

        for bh_type, bh_type_data in bh_data.items():
            if bh_type_data['meta']['count'] > 0:
                # If it's a group we populate members objectId now
                if bh_type == 'groups':
                    await process_members(conn, bh_type_data, group_members, msbh.DNs)
                filectr = filectr_dict.get(bh_type, 0)
                await msbh.write_json_to_zip(bh_type, bh_type_data, filectr)
                LOG.info('Flushing %d %s entries to zip' % (bh_type_data['meta']['count'], bh_type))

    LOG.info('Bloodhound data saved to %s' % msbh.zipfilepath)

async def process_members(conn, bh_type_data, group_members, dn_to_sid):
    """
    For each group entry in bh_type_data, populate its 'Members' property with resolved ObjectIdentifiers.
    Uses dn_to_sid for fast lookup, and LazyAdSchema for unresolved DNs.
    """
    # The dn_to_sid gonna be updated by lazy schema and that's the goal
    global_lazy_adschema.dn_dict = dn_to_sid
    global_lazy_adschema.conn = conn

    # Add all DNs to LazyAdSchema for resolution
    for members_DNs in group_members.values():
        for dn in members_DNs:
            global_lazy_adschema.adddn(dn)

    # For each group entry, populate Members with ObjectIdentifiers
    for entry in bh_type_data['data']:
        group_objid = entry.get('ObjectIdentifier')
        members_DNs = group_members[group_objid]
        members_objids = []
        for dn in members_DNs:
            objid = await global_lazy_adschema.getdn(dn.upper())
            # If no obj id found, skip the member
            if objid:
                if objid.upper() in WELLKNOWN_SIDS:
                    objid = '%s-%s' % (entry['Properties']['domain'], objid.upper())
                members_objids.append({"ObjectIdentifier": objid})
        entry['Members'] = members_objids


import unittest
import inspect
from bloodyAD.cli_modules import msldap


class MSLDAPModuleTests(unittest.TestCase):
    """Tests for the msldap module that exposes MSLDAPClientConsole methods"""
    
    def test_module_has_functions(self):
        """Test that the msldap module has exposed functions"""
        functions = [name for name, obj in inspect.getmembers(msldap, inspect.isfunction)]
        # Should have many functions (around 80+)
        # Note: 7 methods are filtered (ls, cd, rm, nullcb, nocb, bindtree, cat) plus 6 interactive methods
        self.assertGreater(len(functions), 75)
    
    def test_no_interactive_methods_exposed(self):
        """Test that interactive methods are not exposed"""
        functions = [name for name, obj in inspect.getmembers(msldap, inspect.isfunction)]
        
        # These interactive methods should NOT be exposed
        interactive_methods = ['login', 'nosig', 'help', 'quit', 'test', 'plugin']
        for method in interactive_methods:
            self.assertNotIn(method, functions, 
                            f"Interactive method '{method}' should not be exposed")
    
    def test_functions_have_proper_signature(self):
        """Test that exposed functions have proper signatures with 'conn' parameter"""
        # Test a few sample functions
        sample_functions = ['whoami', 'query', 'ldapinfo']
        
        for func_name in sample_functions:
            if not hasattr(msldap, func_name):
                self.fail(f"Function '{func_name}' not found in msldap module")
            
            func = getattr(msldap, func_name)
            sig = inspect.signature(func)
            params = list(sig.parameters.keys())
            
            # First parameter should be 'conn'
            self.assertEqual(params[0], 'conn', 
                           f"First parameter of '{func_name}' should be 'conn'")
    
    def test_functions_have_docstrings(self):
        """Test that exposed functions have docstrings"""
        sample_functions = ['whoami', 'query', 'ldapinfo']
        
        for func_name in sample_functions:
            if not hasattr(msldap, func_name):
                self.fail(f"Function '{func_name}' not found in msldap module")
            
            func = getattr(msldap, func_name)
            self.assertIsNotNone(func.__doc__, 
                               f"Function '{func_name}' should have a docstring")
            self.assertGreater(len(func.__doc__), 0, 
                             f"Function '{func_name}' docstring should not be empty")
    
    def test_functions_have_annotations(self):
        """Test that exposed functions have proper type annotations"""
        sample_functions = ['whoami', 'query']
        
        for func_name in sample_functions:
            if not hasattr(msldap, func_name):
                self.fail(f"Function '{func_name}' not found in msldap module")
            
            func = getattr(msldap, func_name)
            self.assertIn('conn', func.__annotations__, 
                        f"Function '{func_name}' should have 'conn' annotation")
    
    def test_specific_methods_exposed(self):
        """Test that specific important methods are exposed"""
        important_methods = [
            'whoami',
            'query',
            'ldapinfo',
            'user',
            'groupmembers',
            'adduser',
            'deluser',
            'dnsdump'
        ]
        
        for method in important_methods:
            self.assertTrue(hasattr(msldap, method), 
                          f"Method '{method}' should be exposed in msldap module")
    
    def test_parameter_extraction(self):
        """Test that parameters can be extracted correctly for main.py"""
        # This tests the fix for the parameter passing issue
        # main.py extracts parameters using: func.__code__.co_varnames[1:func.__code__.co_argcount]
        
        # Test function with parameters (to_print is now hidden)
        func = getattr(msldap, 'user')
        param_names = func.__code__.co_varnames[1:func.__code__.co_argcount]
        self.assertEqual(param_names, ('samaccountname',),
                        "user function should have samaccountname parameter (to_print is hidden)")
        
        # Test function with no parameters (other than conn)
        func = getattr(msldap, 'whoami')
        param_names = func.__code__.co_varnames[1:func.__code__.co_argcount]
        self.assertEqual(param_names, (),
                        "whoami function should have no parameters")
        
        # Test function with multiple parameters
        func = getattr(msldap, 'query')
        param_names = func.__code__.co_varnames[1:func.__code__.co_argcount]
        self.assertIn('query', param_names,
                     "query function should have query parameter")
        self.assertIn('attributes', param_names,
                     "query function should have attributes parameter")
    
    def test_hidden_parameters_not_exposed(self):
        """Test that show and to_print parameters are not exposed"""
        # Test functions that would normally have these parameters
        test_cases = {
            'user': ['show', 'to_print'],
            'ldapinfo': ['show'],
            'adinfo': ['show'],
        }
        
        for func_name, hidden_params in test_cases.items():
            if hasattr(msldap, func_name):
                func = getattr(msldap, func_name)
                sig = inspect.signature(func)
                param_names = list(sig.parameters.keys())
                
                for hidden_param in hidden_params:
                    self.assertNotIn(hidden_param, param_names,
                                   f"{func_name} should not expose {hidden_param} parameter")
    
    def test_suppressed_methods_not_exposed(self):
        """Test that ls, cd, rm, nullcb, nocb, bindtree, cat are not exposed"""
        suppressed = ['ls', 'cd', 'rm', 'nullcb', 'nocb', 'bindtree', 'cat']
        functions = [name for name, obj in inspect.getmembers(msldap, inspect.isfunction)]
        
        for method in suppressed:
            self.assertNotIn(method, functions,
                           f"{method} should not be exposed in msldap module")


if __name__ == '__main__':
    unittest.main()
